/*
    Copyright (C) 2002-2004 Robert Muth <robert at muth dot org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 2 of June 1991.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program in form of the file COPYING; 
    if not, write to the 

    Free Software Foundation, Inc. <http://www.fsf.org>
    59 Temple Place, Suite 330, 
    Boston, MA 02111-1307  USA

    $Id: basic.H,v 1.4 2004/04/22 21:29:36 muth Exp $
*/

#ifndef BASIC_H
#define BASIC_H

using namespace std;


typedef bool BOOL;
typedef int INT32;
typedef unsigned int UINT32;
typedef float FLT32;
typedef double FLT64;
typedef void VOID;
typedef char CHAR;

#ifndef TRUE
#define TRUE true
#endif

#ifndef FALSE
#define FALSE false
#endif

VOID DisableMessage(UINT32 prefix);
VOID EnableMessage(UINT32 prefix);


extern VOID Message(string prefix, string message, BOOL trace, BOOL abort);
#define Error(message)  do{ ostringstream o; o << message; Message(string("E: "), string(o.str()),FALSE,TRUE);}while(0)
#define Warning(message)  do{ ostringstream o; o << message; Message(string("W: "), string(o.str()),FALSE,FALSE);}while(0)
#define Debug(message)  do{ ostringstream o; o << message; Message(string("D: "), string(o.str()),FALSE,FALSE);}while(0)
#define Information(message)  do{ ostringstream o; o << message; Message(string("I: "), string(o.str()),FALSE,FALSE);}while(0)


extern VOID InstallSignalHandlers(char *name);
extern VOID DumpTrace();
extern VOID DisableDumpTrace();

string StringFromDouble(double val , UINT32 precision, UINT32 width);
string StringDecFromInt32( INT32 l, UINT32 digits=0, CHAR padding=' ');
extern UINT32 Uint32FromString(const string& s);
extern INT32  Int32FromString(const string& s);

string Unquote(const string& is);
string Quote(const string& is);

extern INT32 Tokenize(const string& line, string *array, UINT32 n);

#define ASSERTZ(message) do{ ostringstream o; o << message; Message(string("A: "), string(o.str()),TRUE,TRUE);}while(0)
#define ASSERT(condition,message)   do{ if(!(condition)) ASSERTZ( __FILE__ << ":" << __LINE__ << ": " << message);}while(0)
#define ASSERTX(condition) ASSERT(condition,"assertion failed\n")  

#ifdef __GNUC__
#if __GNUC__ < 3
inline ios& left(ios& i)
{ i.setf(ios::left, ios::adjustfield); return i; }

inline ios& right(ios& i)
{ i.setf(ios::right, ios::adjustfield); return i; }
#endif
#endif


#endif
