/*
    Copyright (c) 2002-2004 Robert Muth <robert at muth dot org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 2 of June 1991.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program in form of the file COPYING; 
    if not, write to the 

    Free Software Foundation, Inc. <http://www.fsf.org>
    59 Temple Place, Suite 330, 
    Boston, MA 02111-1307  USA

    $Id: main.C,v 1.10 2004/04/24 15:37:12 muth Exp $
*/

// ========================================================================

#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <gdk/gdkkeysyms.h>
#include <X11/Xlib.h>

// ========================================================================

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <unistd.h>
#include <pthread.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <dirent.h>

# include <dlfcn.h>

extern "C"
{
#include "xmms/plugin.h"
}

#include <iostream>
#include <string>
#include <fstream>
#include <sstream>
#include <iomanip>

// ========================================================================

#include "basic.H"
#include "switch.H"

// ========================================================================
// GLOBALS
// ========================================================================

const CHAR * Banner =

"# Welcome to Musicus\n"
"# Copyright (c) 2002-2004 Robert Muth  <robert at muth dot org>\n"
"# $Id: main.C,v 1.10 2004/04/24 15:37:12 muth Exp $\n"
"# Entering interactive mode\n";



struct
{
    pthread_mutex_t Mutex;
    string InfoText;
    string Title;
    string Filename;
    INT32 Channels;
    INT32 Length;
    INT32 Rate;
    INT32 Frequency;
    class IPLUGIN *Input;
    class OPLUGIN *Output;    
}
CurrentSong  = 
{
    PTHREAD_MUTEX_INITIALIZER,
    string(""),
    string(""),
    string(""),
    0,
    0,
    0,
    0
};


// ========================================================================

SWITCH SwitchDebug("debug","general",SWITCH_TYPE_BOOL, SWITCH_MODE_OVERWRITE,"0",
                  "enable debug mode");

SWITCH SwitchNoBanner("nobanner","general",SWITCH_TYPE_BOOL, SWITCH_MODE_OVERWRITE,"0",
                      "do not show banner in interactive mode");

SWITCH SwitchLog("log","general",SWITCH_TYPE_STRING, SWITCH_MODE_OVERWRITE,"",
                  "enable debug mode");

SWITCH SwitchPath("path","general",SWITCH_TYPE_STRING, SWITCH_MODE_OVERWRITE,"/usr/lib/xmms/",
                  "plugin path");

SWITCH SwitchOutput("output","general",SWITCH_TYPE_STRING, SWITCH_MODE_OVERWRITE,"libOSS.so",
                  "output plugin");


// ========================================================================
// CALLBACKS (for plugin)
// ========================================================================


string GlobalTitleFormat = "\036p=%p \036t=%t \036a=%a \036n=%n \036y=%y \036d=%d \036g=%g \036c=%c\t";

extern "C"
{


gchar *xmms_get_gentitle_format()
{
    Debug("called xmms_get_gentitle_format()\n");
    
    return const_cast<gchar *>(GlobalTitleFormat.c_str());
}

int effects_enabled(void)
{
    return 0;
}

EffectPlugin *get_current_effect_plugin()
{
    Debug("called get_current_effect_plugin() \n");

    return 0;
}

void ip_add_vis_pcm (int time, AFormat fmt, int nch, int length, void *ptr)
{
    Debug("called ip_add_vis_pcm()\n");
}


void ip_set_info(char *title, int length, int rate, int freq, int nch)
{
    Debug("called ip_set_info()\n");
    
    pthread_mutex_lock(&CurrentSong.Mutex);

    CurrentSong.Title = string(title ? title : "notitle");
    CurrentSong.Frequency = freq;
    CurrentSong.Length = length;
    CurrentSong.Rate = rate;
    CurrentSong.Channels = nch;
    pthread_mutex_unlock(&CurrentSong.Mutex);
}

void ip_set_info_text (char *text)
{
    Debug("called ip_set_info_text()\n");

   pthread_mutex_lock(&CurrentSong.Mutex);

    CurrentSong.InfoText = string(text);

    pthread_mutex_unlock(&CurrentSong.Mutex);
}

}

// ========================================================================
// BASE CLASS PLUGIN
// ========================================================================

typedef void *(*DYNFUNPTR) (void);

class PLUGIN
{
protected:
    void *_handle;
    DYNFUNPTR _gpi;
    string _filename;
    
public:    
    PLUGIN(string& filename, string& symbol) : _filename(filename)
    {
        _handle = dlopen(_filename.c_str(), RTLD_NOW);
        if( _handle == 0 )
        {
            Warning( "cannot open library " << filename << endl);
            
            return;
        }

        _gpi = (DYNFUNPTR) dlsym(_handle,symbol.c_str());

        if( _gpi == 0L ) 
        {
            Warning("cannot find symbol " << symbol << endl);
            
            dlclose(_handle);            
            _handle = 0;
        }
        
    }
    
    
    BOOL IsValid()
    {
        return (_handle != 0);
    }
};

// ========================================================================
// CLASS OUTPUT PLUGIN
// ========================================================================


class OPLUGIN : public PLUGIN
{
private:
    static OPLUGIN *_list;
    static string get_plugin_info;
    
    OPLUGIN *_next;
    OutputPlugin *_interface;
public:
    OPLUGIN(string& filename) : PLUGIN(filename, get_plugin_info)
    {
        if( _handle )
        {
            _interface = reinterpret_cast<OutputPlugin*>( _gpi());
            _interface->handle = _handle;
            _interface->filename = const_cast<gchar *>(_filename.c_str());

            if( _interface->init )
            _interface->init();
            
            _next = _list;
            _list = this;
        }
    }
    
    OutputPlugin *Interface() const
    {
        return _interface;
    }
    
    const gchar *Description() const
    {
        return _interface->description ? _interface->description : "";
    }

    static OPLUGIN *FindPluginByName( const string& name) 
    {
        for( OPLUGIN *ip = _list; ip; ip = ip->_next)
        {
            if( string::npos != ip->_filename.find(name) )
            {
                return ip;
            }
        }

        return 0;
    }

    VOID SetVolume(INT32 lvol, INT32 rvol)
    { 
        if( _interface->set_volume )
        {
            _interface->set_volume(lvol,rvol);
        }
    }    

    VOID GetVolume(INT32& lvol, INT32& rvol)
    { 
        if( _interface->get_volume )
        {
            _interface->get_volume(&lvol,&rvol);
        }
        else
        {
            lvol = 0;
            rvol = 0;
        }
        
    }    

};

// ========================================================================
// CLASS INPUT PLUGIN
// ========================================================================

class IPLUGIN : public PLUGIN
{
private:
    static string get_plugin_info;

    static IPLUGIN *_list;
    
    IPLUGIN *_next;
    InputPlugin *_interface;
public:
    IPLUGIN(string filename) : PLUGIN(filename, get_plugin_info)
    {
        if( _handle )
        {
            _interface = reinterpret_cast<InputPlugin*>( _gpi());
            _interface->handle = _handle;
            _interface->filename = "";//const_cast<gchar *>(_filename.c_str());
            _interface->add_vis_pcm = ip_add_vis_pcm;
            _interface->set_info = ip_set_info;
            _interface->set_info_text = ip_set_info_text;

            if( _interface->init )
                _interface->init();

            _next = _list;
            _list = this;
        }
    }
    

    const gchar *Description() const
    {
        return _interface->description ? _interface->description : "";
    }
    
    VOID PlaySong(gchar * song, OPLUGIN *op)
    {
        _interface->output = op->Interface();
        _interface->play_file( song );
    }

    VOID StopSong()
    {
        if( _interface->stop )
        _interface->stop();
    }

    VOID PauseSong()
    {
        if( _interface->pause )
            _interface->pause(1);
    }

    VOID UnpauseSong()
    {
        if( _interface->pause )
            _interface->pause(0);
    }

    string GetInfo(string filename)
    {
        if( _interface->get_song_info )
        {
            char *info;
            INT32 length;

            // we may have a memory leak here - investigate
            
            _interface->get_song_info(const_cast<char *>(filename.c_str()),&info,&length);

            return  StringDecFromInt32(length) + " " + info;
        }
        else
        {
            return string("0 @");
        }
    }




    INT32 GetTime()
    {
        if( _interface->get_time )
        {
            return _interface->get_time();
        }
        else
        {
            return -1;
        }
    }
    
    VOID SeekTime(INT32 time)
    {
        if( _interface->seek )
        {
            _interface->seek(time);
        }
    }
    
    
    static IPLUGIN *FindPluginForSong( const gchar *songname) 
    {
        for( IPLUGIN *ip = _list; ip; ip = ip->_next)
        {
            if( ip->_interface->is_our_file(const_cast<gchar *>(songname)) )
            {
                return ip;
            }
        }

        return 0;
    }

    
};

// ========================================================================

OPLUGIN *OPLUGIN::_list = 0;
string OPLUGIN::get_plugin_info = string("get_oplugin_info");
IPLUGIN *IPLUGIN::_list = 0;
string IPLUGIN::get_plugin_info = string("get_iplugin_info");

// ========================================================================
 


//void xmms_usleep(gint usec);



// ========================================================================

bool FileOk( string filename)
{
    struct stat statbuf;

    // there are some issues with this plugin 
    if(  filename.find("libidcin.so") != string::npos ) return false;
    
    
    if( stat(filename.c_str(), &statbuf) ) return false;
    
    
    if( (statbuf.st_mode & S_IFMT) == S_IFLNK ) return false;


    if( (statbuf.st_mode & S_IFMT) != S_IFREG ) return false;
    
    return true;
}


void LoadPlugins()
{
    string root = SwitchPath.ValueString();
    DIR *directory;    
    string path;

    path = root + "Output";
    directory = opendir(path.c_str());
    if ( directory)
    {
        struct dirent *ent;

        while ((ent = readdir(directory)) != 0)
        {
            string filename = path + "/" +  ent->d_name;

            if( FileOk(filename) )
            {
                OPLUGIN *op = new OPLUGIN(filename);
                if( !op->IsValid() )
                {
                    delete op;
                }
                else
                {
                    Information("READ " << filename << ": " << op->Description() << endl);
                }
                
            }
        }
    }

    path = root + "Input";
    directory = opendir(path.c_str());
    if ( directory)
    {
        struct dirent *ent;

        while ((ent = readdir(directory)) != 0)
        {
            string filename = path + "/" +  ent->d_name;

            if( FileOk(filename) )
            {
                IPLUGIN *ip = new IPLUGIN(filename);

                if( !ip->IsValid() )
                {
                    delete ip;
                }
                else
                {
                    Information("READ " << filename << ": " << ip->Description() << endl);
                }

            }
        }
    }

}

// ========================================================================

VOID HelpMessage(ostream &out)
{
    out << 
        "Interactive Commands: \n"
        "\n"
        "help                  print this page\n"
        "version               print version\n"
        "debug                 enable debug messages\n"
        "plugins               print available plugins\n"
        "quit                  exit program\n"
        "\n"
        "setvol <lvol> <rvol>  set left and right volume [0-100]\n"
        "getvol                get current volume\n"
        "\n"
        "play <filename>       play song <filename>\n"  
        "stop                  stop song\n"
        "pause                 pause song\n"
        "unpause               unpause song\n"
        "setpos <pos>          move to song pos [msec]\n"       
        "getpos                get current song position [msec]\n"       
        "getlength             get song length  [msec]\n"
        "settitleformat        set title string format\n"
        "getinfocurr           get current song information (length and title string)\n"
        "getinfo <filename>    get song information (length and title string)\n";
    
}


// ========================================================================

ofstream *GlobalLog = new ofstream("/dev/null", ios::out);



// ========================================================================

VOID UserOk(ostream &out, const string& type)
{
    out << "@ OK " << type << "\n";
    *GlobalLog << "@ OK " << type << "\n";
}

// ========================================================================

VOID UserOk(ostream &out, const string& type, const string& message)
{
    out << "@ OK " << type << " " << Quote(message) << "\n";
    *GlobalLog << "@ OK " << type << " " << Quote(message) << "\n";
}

// ========================================================================

VOID UserOk(ostream &out, const string& type, const string& filename, const string& message)
{
    out << "@ OK " << type << " " << Quote(filename) << " " << Quote(message) << "\n";
    *GlobalLog << "@ OK " << type <<  " " << Quote(filename) << " " << Quote(message) << "\n";
}

// ========================================================================

VOID UserOk(ostream &out, const string type, INT32 a,INT32 b)
{
    out << "@ OK " << type << " " << a << " " << b  << "\n";
    *GlobalLog << "@ OK " << type << " " << a << " " << b  << "\n";
}

// ========================================================================

VOID UserOk(ostream &out, const string type, INT32 a)
{
    out << "@ OK " << type << " " <<  a << "\n";
    *GlobalLog<< "@ OK " << type << " " <<  a << "\n";
}

// ========================================================================

VOID UserError(ostream &out,const string type, const string& message)
{
    out << "@ ERROR " << type << " " << Quote(message) << "\n";
    *GlobalLog << "@ ERROR " << type << " " << Quote(message) << "\n";
}

// ========================================================================

VOID UserError(ostream &out,const string type, const string& filename, const string& message)
{
    out << "@ ERROR " << type << " " << Quote(filename) << " " << Quote(message) << "\n";
    *GlobalLog << "@ ERROR " << type << " " << Quote(filename) << " " << Quote(message) << "\n";
}

// ========================================================================

VOID UserInput(ostream &out,const string& message)
{
    *GlobalLog << message << endl;
}

// ========================================================================

VOID *InteractiveMode(VOID *x, ostream &out,istream &inp)
{

    if( !SwitchNoBanner.ValueBool() )
    {
        out << Banner;
        
            
    }
    
    Information("Entering Interactive mode\n");
    
    while(1)
    {
        string line;
 
        getline(inp, line);

        if (inp.eof())
        {
            // Control-D
            break;
        }


        UserInput(out,line);

        string inputs[16];

        const UINT32 n =  Tokenize(line, inputs, 16);

#if 0
        for( UINT32 i=0; i<n; ++i )
        {
            out << i << ">>>" << inputs[i] << "<<<\n";
        }
#endif   
                
        if( inputs[0] == "help" )
        {
            HelpMessage(cout);
            UserOk(out, inputs[0]);
        }   

        // ------------------------------------------------------------------------------------------------

        else if( inputs[0] == "quit" )
        {
            UserOk(out, inputs[0]);
            break;
        }

        // ------------------------------------------------------------------------------------------------

        else if( inputs[0] == "debug" )
        {
            EnableMessage('I');
            //EnableMessage('D');
            //EnableMessage('W');
            UserOk(out,inputs[0]);
        }

        // ------------------------------------------------------------------------------------------------

        else if( inputs[0] == "version" )
        {
            UserOk(out,inputs[0],"$Id: main.C,v 1.10 2004/04/24 15:37:12 muth Exp $");
        }

        // ------------------------------------------------------------------------------------------------

        else if( inputs[0] == "plugins" )
        {
            UserError(out,inputs[0],"not yet implemented");
        }

        // ------------------------------------------------------------------------------------------------

        else if( inputs[0] == "setvol" )
        {

            if( n <= 2 )
            {
                UserError(out, inputs[0], "parameter missing");
            }
            else
            {
                const INT32 lvol = Int32FromString( inputs[1] );
                const INT32 rvol = Int32FromString( inputs[2] );
                
                if( lvol < 0 || lvol > 100 )
                {
                    UserError(out, inputs[0], "pos out of range");
                }
                else if( rvol < 0 || rvol > 100 )
                {
                    UserError(out, inputs[0], "pos out of range");
                }
                else
                {
                    CurrentSong.Output->SetVolume(lvol,rvol);
                    UserOk(out, inputs[0]);
                }
            }
            
        }

        // ------------------------------------------------------------------------------------------------

        else if( inputs[0] == "getvol" )
        {
        
            INT32 lvol;
            INT32 rvol;
            CurrentSong.Output->GetVolume(lvol,rvol);
            UserOk(out, inputs[0],lvol,rvol);
        }
        
        
        // ------------------------------------------------------------------------------------------------

        else if( inputs[0] == "play" )
        {
            if( CurrentSong.Input != 0 )
            {
                UserError(out,inputs[0],"song already playing");
            }
            else if ( n <= 1 )
            {
                UserError(out, inputs[0], "parameter missing");
            }
            else
            {
                CurrentSong.Filename = Unquote(inputs[1]);
                
                CurrentSong.Input  = IPLUGIN::FindPluginForSong(CurrentSong.Filename.c_str());
                if( CurrentSong.Input )
                {
                    CurrentSong.Input->PlaySong( const_cast<gchar *>(CurrentSong.Filename.c_str()), CurrentSong.Output );
                    UserOk(out, inputs[0], CurrentSong.Input->Description() );
                }
                else
                {
                    UserError(out, inputs[0], "cannot play song");
                }
                
            }
        }

        // ------------------------------------------------------------------------------------------------

        else if( inputs[0] == "getinfo" )
        {
            IPLUGIN *input  = IPLUGIN::FindPluginForSong(inputs[1].c_str());
            if( input )
            {
                string  s = input->GetInfo( Unquote(inputs[1]) );
                UserOk(out, inputs[0], inputs[1], s);    // NB: sone is part of output
            }
            else if ( n <= 1 )
            {
                UserError(out, inputs[0], "parameter missing");   // this is inconsistent with providing the song as a part of the put 
            }
            else
            {
                UserError(out, inputs[0], "no suitable plugin found");
            }
        }

        // ------------------------------------------------------------------------------------------------

        else if( inputs[0] == "settitleformat" )
        {
            if ( n <= 1 )
            {
                UserError(out, inputs[0], "parameter missing");
            }
            else
            {
                GlobalTitleFormat = Unquote(inputs[1]);
            
                UserOk(out, inputs[0]);
            }
        }

        // ------------------------------------------------------------------------------------------------

        else if( inputs[0] == "stop" )
        {
            if( CurrentSong.Input == 0 )
            {
                UserError(out,inputs[0],"no song playing");
            }
            else
            {
                CurrentSong.Input->StopSong();
                CurrentSong.Input = 0;
                UserOk(out, inputs[0]);
            }
            
        }

        // ------------------------------------------------------------------------------------------------

        else if( inputs[0] == "pause" )
        {
            if( CurrentSong.Input == 0 )
            {
                UserError(out, inputs[0],"no song playing");
            }
            else
            {
                CurrentSong.Input->PauseSong();
                UserOk(out, inputs[0]);
            }
            
        }

        // ------------------------------------------------------------------------------------------------

        else if( inputs[0] == "unpause" )
        {
            if( CurrentSong.Input == 0 )
            {
                UserError(out,inputs[0],"no song playing");
            }
            else
            {
                CurrentSong.Input->UnpauseSong();
                UserOk(out, inputs[0]);
            }
            
        }

        // ------------------------------------------------------------------------------------------------

        else if( inputs[0] == "getpos" )
        {
            if( CurrentSong.Input == 0 )
            {
                UserError(out,inputs[0],"no song playing");
            }
            else
            {
                INT32 time = CurrentSong.Input->GetTime();
                UserOk(out, inputs[0], time);
            }
            
        }

        // ------------------------------------------------------------------------------------------------
	
	else if( inputs[0] == "getinfocurr" )
        {
            if( CurrentSong.Input == 0 )
            {
                UserError(out,inputs[0],"no song playing");
            }
            else
            {
                string  s = CurrentSong.Input->GetInfo(CurrentSong.Filename);
                UserOk(out, inputs[0], Quote(CurrentSong.Filename), s);
            }
	}

    	// ------------------------------------------------------------------------------------------------
    
        else if( inputs[0] == "setpos" )
        {
            if( CurrentSong.Input == 0 )
            {
                UserError(out,inputs[0],"no song playing");
            }
            else if ( n <= 1 )
            {
                UserError(out, inputs[0], "parameter missing");
            }
            else 
            {
                const INT32 pos = Int32FromString( inputs[1] );
                
                if( pos < 0 || pos > CurrentSong.Length )
                {
                    UserError(out,inputs[0],"pos out of range");
                }
                else
                {
                    // this is insane,can only seek with sec granularity
                    CurrentSong.Input->SeekTime(pos/1000);
                    
                    UserOk(out,inputs[0],pos);
                }
                
            }
            
        }

        // ------------------------------------------------------------------------------------------------

        else if( inputs[0] == "getlength" )
        {
            if( CurrentSong.Input == 0 )
            {
                UserError(out,inputs[0],"no song playing");
            }
            else
            {
                INT32 time = CurrentSong.Length;

                UserOk(out,inputs[0],time);
            }
            
        }

        // ------------------------------------------------------------------------------------------------

        else
        {
            UserError(out,inputs[0],"unknown command");
            //break;
        }
        
    }

    return 0;
}


// ========================================================================
// ========================================================================


int main(int argc,char *argv[])
{
    //

    InstallSignalHandlers(argv[0]);
    
    //    ParseCommandLine

    for( argc--, argv++; argc > 0;  argc--, argv++)
    {
        if( *argv[0] == '-' )
        {
            SWITCH  *sw = SWITCH::SwitchFind(argv[0]+1);
            if( sw == 0 )
            {
                Error("bad option " << argv[0] << endl);
            }

            if( sw->Type() == SWITCH_TYPE_BOOL )
            {
                sw->ValueAdd("1");
            }
            else
            {
                ASSERTX( argc > 0);
                argc--;
                argv++;
                sw->ValueAdd(argv[0]);
            }
        }
        else
        {
            break;
        }
        
    }

    if( !SwitchDebug.ValueBool() )
    {
        DisableMessage('I');
        DisableMessage('D');
        DisableMessage('W');
    }


    // Loading of Plugins

    LoadPlugins();

    // Set Output

    CurrentSong.Output = OPLUGIN::FindPluginByName( SwitchOutput.ValueString().c_str());
    if( CurrentSong.Output )
    {
        if( SwitchDebug.ValueBool() )
        {
            Information("found " << CurrentSong.Output->Description() << endl);
        }
        
    }
    else
    {
        Error("could not find output plugin " << SwitchOutput.ValueString() << endl);
        return -1;
    }


#if 0
    pthread_t thread_handle;
    pthread_create(&thread_handle,NULL,InteractiveMode,NULL);
#endif

    if( argc == 0 )
    {
        if( SwitchLog.ValueString() != "" )
        {
            delete GlobalLog;
            
            GlobalLog = new ofstream( SwitchLog.ValueString().c_str(), ios::out);
        }

        InteractiveMode(NULL,cout,cin);
    }
    else
    {
        for( ; argc > 0;  argc--, argv++)
        {
            Information(" playing "<< argv[0] << endl);
            CurrentSong.Input  = IPLUGIN::FindPluginForSong(argv[0]);
            if( CurrentSong.Input )
            {
                Information( "found " << CurrentSong.Input->Description() << endl);
                CurrentSong.Input->PlaySong( argv[0], CurrentSong.Output );
                
                while(1)
                {
                    INT32 time = CurrentSong.Input->GetTime();
                    sleep(1);
                    
                    if( time < 0 ) 
                    {
                        CurrentSong.Input->StopSong();
                        break;
                    }
                    
                    
                    Information(CurrentSong.Title << " " << StringFromDouble( FLT64(time) / FLT64(CurrentSong.Length),3,5) << endl);
                }
                
            }
            else
            {
                cout << "no player found\n";
            }
        }
    }
    
    return 0;
}
        
// ========================================================================
